# Unsloth Studio
# Copyright (C) 2024-present the Unsloth AI team. All rights reserved.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from typing import Optional, Tuple, Dict, List, Callable, Any
from torch._dynamo import mark_dynamic

UNSLOTH_COMPILE_ENABLE  = os.environ.get("UNSLOTH_COMPILE_DISABLE", "0") == "0"
UNSLOTH_COMPILE_DEBUG   = os.environ.get("UNSLOTH_COMPILE_DEBUG",   "0") == "1"
UNSLOTH_COMPILE_MAXIMUM = os.environ.get("UNSLOTH_COMPILE_MAXIMUM", "0") == "1"
torch_compile_options = {
    "epilogue_fusion"   : True,
    "max_autotune"      : UNSLOTH_COMPILE_MAXIMUM,
    "shape_padding"     : True,
    "trace.enabled"     : UNSLOTH_COMPILE_DEBUG,
    "triton.cudagraphs" : False,
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "test_efficient_ce_loss",
]
from ..utils import *
from ..losses import (
    UnslothEfficientLoss,
    unsloth_efficient_ce_loss,
)
import torch


def _test_efficient_ce_loss(
    bsz = 4,
    qlen = 2009,
    hd = 4096,
    vocab_size = 16 * 1024,
    dtype = torch.float16,
    reduction = "sum",
    logit_scale = None,
    logit_softcapping = None,
    random_state = 3407,
    has_bias = True,
    weight_requires_grad = True,
    bias_requires_grad   = True,
    ignore_index = -100,
    device = "cuda",
):
    # All Unsloth Studio code licensed under AGPLv3
    torch.cuda.manual_seed(random_state)
    torch.manual_seed(random_state)
    if not has_bias: bias_requires_grad = False

    # Increase atol / rtol for float16 due to float16_scaler being 65536
    # https://pytorch.org/docs/stable/testing.html#torch.testing.assert_close
    if dtype == torch.float16:
        rtol = 1e-3 * 65536 / 100
        atol = 1e-5 * 65536 / 10
    else:
        rtol = 1.6e-2
        atol = 1e-5

    # Use grad scaling for float16 as well!
    scaler = torch.amp.GradScaler() if dtype == torch.float16 else None
    hidden_states = torch.randn((bsz, qlen, hd), dtype = dtype, requires_grad = True, device = device)
    if weight_requires_grad or bias_requires_grad:
        lm_head = torch.nn.Linear(hd, vocab_size, bias = has_bias, device = device, dtype = torch.float32)
    else:
        # Downcast to bfloat16
        lm_head = torch.nn.Linear(hd, vocab_size, bias = has_bias, device = device, dtype = dtype)
    lm_head.weight.requires_grad_(weight_requires_grad)
    if has_bias: lm_head.bias.requires_grad_(bias_requires_grad)
    labels = torch.randint(0, vocab_size, (bsz, qlen), device = device).to(torch.int64)
    padding = torch.randint(0, int(qlen * 0.2), (bsz, ), device = device)
    for i in range(bsz):
        labels[i, padding[i]:] = ignore_index

    n_items = (labels != ignore_index).sum()

    # Get original CE Loss
    hidden_states.grad = None
    lm_head.weight.grad = None
    if has_bias: lm_head.bias.grad = None

    with torch.amp.autocast(device_type = "cuda", dtype = dtype):
        logits = lm_head(hidden_states)

    with torch.autocast(device_type = "cuda", enabled = False):
        logits = logits.float()
        # Logit Scaling like in Cohere, Granite
        if logit_scale is not None:
            logits = logits * logit_scale
        # Softcapping like in Gemma 2
        if logit_softcapping is not None:
            logits = logits / logit_softcapping
            logits = torch.tanh(logits)
            logits = logits * logit_softcapping

        shift_logits = logits[..., :-1, :].contiguous()
        shift_labels = labels[..., 1:].contiguous()
        loss_fct = torch.nn.CrossEntropyLoss(
            reduction = reduction,
            ignore_index = ignore_index,
        )
        shift_logits = shift_logits.view(-1, vocab_size)
        shift_labels = shift_labels.view(-1)
        shift_labels = shift_labels.to(shift_logits.device)
        loss = loss_fct(shift_logits, shift_labels)
        loss = loss / n_items
        old_loss = loss.detach()
    pass
    if scaler is not None: loss = scaler.scale(loss)
    loss.backward()
    old_hidden_states_grad = hidden_states.grad.detach()

    old_weight_grad = lm_head.weight.grad if lm_head.weight is not None else None
    old_bias_grad = lm_head.bias.grad if lm_head.bias is not None else None
    if old_weight_grad is not None: old_weight_grad = old_weight_grad.detach()
    if old_bias_grad is not None: old_bias_grad = old_bias_grad.detach()
    print(
        loss.item(),
        torch.amax(old_hidden_states_grad).item(),
        torch.amax(old_weight_grad).item() if old_weight_grad is not None else None,
        torch.amax(old_bias_grad).item() if old_bias_grad is not None else None
    )

    # Get new CE Loss
    hidden_states.grad = None
    lm_head.weight.grad = None
    if has_bias: lm_head.bias.grad = None

    with torch.amp.autocast(device_type = "cuda", dtype = dtype):

        loss = unsloth_efficient_ce_loss(
            hidden_states = hidden_states,
            lm_head = lm_head,
            labels = labels,
            shift = True,
            reduction = reduction,
            logit_scale = logit_scale,
            logit_softcapping = logit_softcapping,
            ignore_index = ignore_index,
            chunk_size = 512,
        )
        loss = loss / n_items
        new_loss = loss.detach()
    pass
    if scaler is not None: loss = scaler.scale(loss)
    loss.backward()
    torch.testing.assert_close(new_loss, old_loss, atol = 0.1, rtol = 1e-2)
    new_hidden_states_grad = hidden_states.grad.detach()
    new_weight_grad = lm_head.weight.grad if lm_head.weight is not None else None
    new_bias_grad = lm_head.bias.grad if lm_head.bias is not None else None
    if new_weight_grad is not None: new_weight_grad = new_weight_grad.detach()
    if new_bias_grad is not None: new_bias_grad = new_bias_grad.detach()
    print(
        loss.item(),
        torch.amax(new_hidden_states_grad).item(),
        torch.amax(new_weight_grad).item() if new_weight_grad is not None else None,
        torch.amax(new_bias_grad).item() if new_bias_grad is not None else None
    )

    torch.testing.assert_close(new_hidden_states_grad, old_hidden_states_grad, atol = atol, rtol = rtol)
    if weight_requires_grad:
        assert(new_weight_grad is not None and old_weight_grad is not None)
        torch.testing.assert_close(new_weight_grad, old_weight_grad, atol = atol, rtol = rtol)
    else:
        assert(new_weight_grad is None and old_weight_grad is None)

    if bias_requires_grad:
        assert(new_bias_grad is not None and old_bias_grad is not None)
        torch.testing.assert_close(new_bias_grad, old_bias_grad, atol = atol / 100, rtol = rtol / 100)
    else:
        assert(new_bias_grad is None and old_bias_grad is None)
pass


def test_efficient_ce_loss():
    _test_efficient_ce_loss(
        bsz = 4,
        qlen = 2048,
        hd = 4096,
        vocab_size = 16 * 1024,
        dtype = torch.float16,
        reduction = "sum",
        logit_scale = None,
        logit_softcapping = None,
        random_state = 3407,
        has_bias = True,
        weight_requires_grad = True,
        bias_requires_grad   = True,
        ignore_index = -100,
        device = "cuda",
    )
    torch.cuda.empty_cache()
    _test_efficient_ce_loss(
        bsz = 3,
        qlen = 2047,
        hd = 4097,
        vocab_size = 64 * 1024,
        dtype = torch.bfloat16,
        reduction = "sum",
        logit_scale = None,
        logit_softcapping = 50.0,
        random_state = 3409,
        has_bias = False,
        weight_requires_grad = False,
        bias_requires_grad   = False,
        ignore_index = -101,
        device = "cuda",
    )
    torch.cuda.empty_cache()
    _test_efficient_ce_loss(
        bsz = 1,
        qlen = 1023,
        hd = 2048,
        vocab_size = 128 * 1024,
        dtype = torch.float16,
        reduction = "mean",
        logit_scale = 0.125,
        logit_softcapping = None,
        random_state = 3410,
        has_bias = False,
        weight_requires_grad = True,
        bias_requires_grad   = True,
        ignore_index = -200,
        device = "cuda",
    )
    torch.cuda.empty_cache()
    _test_efficient_ce_loss(
        bsz = 1,
        qlen = 512,
        hd = 2000,
        vocab_size = 256 * 1024,
        dtype = torch.bfloat16,
        reduction = "mean",
        logit_scale = 0.125,
        logit_softcapping = 50.0,
        random_state = 3411,
        has_bias = True,
        weight_requires_grad = True,
        bias_requires_grad   = True,
        ignore_index = -10,
        device = "cuda",
    )
    torch.cuda.empty_cache()
pass



# Unsloth Studio
# Copyright (C) 2024-present the Unsloth AI team. All rights reserved.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from IPython.display import clear_output
import subprocess
import os
os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"
os.environ["GRADIO_ANALYTICS_ENABLED"] = "False"
MODEL_NAME = "unsloth/Meta-Llama-3.1-8B-Instruct-bnb-4bit"

print("Installing packages for 🦥 Unsloth Studio ... Please wait 5 minutes ...")

install_first = [
    "pip", "install",
    "unsloth"
]

install_first = subprocess.Popen(install_first)
install_first.wait()

uninstall_install_first = [
    "pip", "uninstall", "-y",
    "unsloth"
]

uninstall_install_first = subprocess.Popen(uninstall_install_first)
uninstall_install_first.wait()

install_second = [
    "pip", "install",
    "gradio==4.44.1",
    "unsloth[colab-new]@git+https://github.com/unslothai/unsloth.git",
]
install_second = subprocess.Popen(install_second)
install_second.wait()

upgrade_huggingface_hub = [
    "pip", "install", 
    "huggingface-hub",
    "-U",
    "--force-reinstall"
]

upgrade_huggingface_hub = subprocess.Popen(upgrade_huggingface_hub)
upgrade_huggingface_hub.wait()


from huggingface_hub import snapshot_download
import warnings
warnings.filterwarnings(action = "ignore", category = UserWarning,    module = "torch")
warnings.filterwarnings(action = "ignore", category = UserWarning,    module = "huggingface_hub")
warnings.filterwarnings(action = "ignore", category = FutureWarning,  module = "huggingface_hub")
warnings.filterwarnings(action = "ignore", category = RuntimeWarning, module = "subprocess")
warnings.filterwarnings(action = "ignore", category = UserWarning,    module = "transformers")
warnings.filterwarnings(action = "ignore", category = FutureWarning,  module = "accelerate")
warnings.filterwarnings(action = "ignore", category = RuntimeWarning, module = "multiprocessing")
warnings.filterwarnings(action = "ignore", category = RuntimeWarning, module = "multiprocess")

from huggingface_hub.utils import disable_progress_bars
disable_progress_bars()
snapshot_download(repo_id = MODEL_NAME, repo_type = "model")

clear_output()


from contextlib import redirect_stdout
import io
import logging
logging.getLogger("transformers.utils.hub").setLevel(logging.CRITICAL+1)

print("Loading model ... Please wait 1 more minute! ...")

with redirect_stdout(io.StringIO()):
    from unsloth import FastLanguageModel
    import torch
    model, tokenizer = FastLanguageModel.from_pretrained(
        model_name = MODEL_NAME,
        max_seq_length = None,
        dtype = None,
        load_in_4bit = True,
    )
    FastLanguageModel.for_inference(model)
pass
clear_output()

import gradio
gradio.strings.en["SHARE_LINK_DISPLAY"] = ""
from transformers import TextIteratorStreamer, StoppingCriteria, StoppingCriteriaList
from threading import Thread

class StopOnTokens(StoppingCriteria):
    def __init__(self, stop_token_ids):
        self.stop_token_ids = tuple(set(stop_token_ids))
    pass

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
        return input_ids[0][-1].item() in self.stop_token_ids
    pass
pass

def async_process_chatbot(message, history):
    eos_token = tokenizer.eos_token
    stop_on_tokens = StopOnTokens([eos_token,])
    text_streamer  = TextIteratorStreamer(tokenizer, skip_prompt = True)

    # From https://www.gradio.app/guides/creating-a-chatbot-fast
    history_transformer_format = history + [[message, ""]]
    messages = []
    for item in history_transformer_format:
        messages.append({"role": "user",      "content": item[0]})
        messages.append({"role": "assistant", "content": item[1]})
    pass
    # Remove last assistant and instead use add_generation_prompt
    messages.pop(-1)

    input_ids = tokenizer.apply_chat_template(
        messages,
        add_generation_prompt = True,
        return_tensors = "pt",
    ).to("cuda", non_blocking = True)

    # Add stopping criteria - will not output EOS / EOT
    generation_kwargs = dict(
        input_ids = input_ids,
        streamer = text_streamer,
        max_new_tokens = 1024,
        stopping_criteria = StoppingCriteriaList([stop_on_tokens,]),
        temperature = 0.7,
        do_sample = True,
    )
    thread = Thread(target = model.generate, kwargs = generation_kwargs)
    thread.start()

    # Yield will save the output to history!
    generated_text = ""
    for new_text in text_streamer:
        if new_text.endswith(eos_token):
            new_text = new_text[:len(new_text) - len(eos_token)]
        generated_text += new_text
        yield generated_text
    pass
pass

studio_theme = gradio.themes.Soft(
    primary_hue = "teal",
)

scene = gradio.ChatInterface(
    async_process_chatbot,
    chatbot = gradio.Chatbot(
        height = 325,
        label = "Unsloth Studio Chat",
    ),
    textbox = gradio.Textbox(
        placeholder = "Message Unsloth Chat",
        container = False,
    ),
    title = None,
    theme = studio_theme,
    examples = None,
    cache_examples = False,
    retry_btn = None,
    undo_btn = "Remove Previous Message",
    clear_btn = "Restart Entire Chat",
)

scene.launch(quiet = True)



# Unsloth Studio
# Copyright (C) 2024-present the Unsloth AI team. All rights reserved.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "calculate_mmlu",
]

import os
import torch
import numpy as np
import pandas as pd
from datasets import load_dataset
import functools
from tqdm import tqdm


@functools.lru_cache(2)
def get_mmlu_dataset(
    n_samples = 200,
    random_state = 3407,
    add_space = False, # Adds space between question & answer
):
    mmlu_dataset = load_dataset("unsloth/studio_mmlu", split = "train")

    # Only used if tokenizer does not use _A, _B, _C, _D, and just uses
    # A, B, C, D. Llama 3.1 has _A.
    space = " " if add_space else ""

    def mmlu_tokenize(tokenizer):
        def _mmlu_tokenize(examples):
            all_shots = [
                examples["5_shot"],
                examples["4_shot"],
                examples["3_shot"],
                examples["2_shot"],
                examples["1_shot"],
            ]
            question = examples["Q"]
            prompts = []
            for shot, q in zip(all_shots[0], question):
                prompts.append(shot + space + q)
            input_ids = tokenizer(prompts).input_ids

            # Original MMLU https://github.com/hendrycks/test
            # uses 2048 as maximum sequence length
            # then reduces 5 shot to 4 shot etc
            for j, input_ids_ in enumerate(input_ids):
                if len(input_ids_) <= 2048: continue
                shot = 0
                while len(input_ids_) > 2048:
                    shot += 1
                    prompt = all_shots[shot][j] + q
                    input_ids_ = tokenizer(prompt).input_ids
                pass
                input_ids[j] = input_ids_
            pass
            return { "input_ids" : input_ids, }
        pass
        return _mmlu_tokenize
    pass
    mmlu_dataset = mmlu_dataset.map(mmlu_tokenize(tokenizer), batched = True, num_proc = 4,)

    # We then sort by lengths of input_ids for faster processing
    mmlu_dataset = mmlu_dataset.map(
        lambda xs: {"lengths" : list(map(len, xs["input_ids"]))},
        batched = True, num_proc = 4,
    )
    mmlu_dataset = mmlu_dataset.to_pandas()

    subjects = mmlu_dataset["Subject"]
    counts = subjects.value_counts()

    # Sklearn type stratified sampling weights
    weights = len(subjects) / (len(counts) * counts)
    mmlu_dataset = mmlu_dataset.sample(n_samples, random_state = random_state, weights = subjects.map(weights))
    mmlu_dataset.reset_index(drop = True, inplace = True)

    sorted_indices = mmlu_dataset["input_ids"].str.len().sort_values().index.values

    return mmlu_dataset, sorted_indices
pass


@functools.lru_cache(2)
def prepare_calculate_mmlu(
    tokenizer,
    n_samples = 200,
    random_state = 3407,
    target_length = 4096,
):
    # Llama 3.1 tokenizer uses _A and not A
    # We shall check this for the specific tokenizer!
    A = tokenizer(" A", add_special_tokens = False).input_ids
    B = tokenizer(" B", add_special_tokens = False).input_ids
    C = tokenizer(" C", add_special_tokens = False).input_ids
    D = tokenizer(" D", add_special_tokens = False).input_ids
    if len(A) == 1 and len(B) == 1 and len(C) == 1 and len(D) == 1:
        answer_ids = torch.tensor([A[0], B[0], C[0], D[0]])
        answer_ids = answer_ids.to("cuda", non_blocking = True)
        mmlu_dataset, sorted_indices = get_mmlu_dataset(n_samples = n_samples, random_state = random_state, add_space = False)
    else:
        A = tokenizer("A", add_special_tokens = False).input_ids
        B = tokenizer("B", add_special_tokens = False).input_ids
        C = tokenizer("C", add_special_tokens = False).input_ids
        D = tokenizer("D", add_special_tokens = False).input_ids
        # We need to add a space since _A does not exist!
        mmlu_dataset, sorted_indices = get_mmlu_dataset(n_samples = n_samples, random_state = random_state, add_space = True)
    pass
    all_lengths = np.fromiter(mmlu_dataset["lengths"], dtype = int)
    
    return mmlu_dataset, sorted_indices, answer_ids, all_lengths
pass


@torch.inference_mode
def calculate_mmlu(
    model,
    tokenizer,
    n_samples = 200,
    random_state = 3407,
    target_length = 4096,
):
    """
    Calculates Macro MMLU Average with Stratified Random Sampling like in
    Tiny MMLU. Original MMLU from https://github.com/hendrycks/test.
    """
    answers_mapping = {"A" : 0, "B" : 1, "C" : 2, "D" : 3,}
    mmlu_dataset, sorted_indices, answer_ids, all_lengths = prepare_calculate_mmlu(
        tokenizer = tokenizer,
        n_samples = n_samples,
        random_state = random_state,
        target_length = target_length,
    )
    n_examples = len(mmlu_dataset)

    n_samples = 0
    i = 0
    total_correct = torch.zeros(1, dtype = int, device = "cuda")

    with torch.amp.autocast(device_type = "cuda", dtype = model.config.torch_dtype), tqdm(total = n_examples) as progress_bar:
        while i < n_examples:
            # Find samples until (target_length = 4096) is reached
            length = 0
            left = i
            while length < target_length:
                index = sorted_indices[i]
                length += all_lengths[index]
                i += 1
                if i == n_examples: break
            right = i
            right = min(right, n_examples)
            i = right
            n_samples += (right-left)
            progress_bar.update(right-left)

            # Select batch
            indices = sorted_indices[left:right]
            current_batch = mmlu_dataset.iloc[indices]
            input_ids = current_batch["input_ids"]
            answers   = current_batch["A"]
            answers   = torch.from_numpy(
                np.fromiter((answers_mapping[x] for x in answers), dtype = int),
            ).to("cuda", non_blocking = True)

            # Pad to longest item
            lengths = [len(x) for x in input_ids]
            max_length = max(lengths)
            paddings = np.fromiter((max_length-len(x) for x in input_ids), dtype = int)
            indexing = max_length - paddings - 1
            input_ids = torch.from_numpy(
                np.vstack([np.hstack((input_ids_, np.zeros(padding, dtype = int))) \
                for input_ids_, padding in zip(input_ids, paddings)])
            ).to("cuda", non_blocking = True)

            logits = model(
                input_ids = input_ids,
                output_attentions = False,
                output_hidden_states = False,
                use_cache = False,
            ).logits

            # As per original MMLU, select A, B, C, D locations
            logits = logits[range(input_ids.shape[0]), indexing, :][:, answer_ids]

            # Original MMLU argmax - careful of ties!
            is_correct = logits.argmax(-1) == answers
            total_correct += is_correct.sum()

            progress_bar.set_description("MMLU Accuracy = {:.1f}%".format(total_correct.item() / n_samples * 100))
        pass
    pass
    return total_correct.item() / n_samples
pass



# Unsloth Studio
# Copyright (C) 2024-present the Unsloth AI team. All rights reserved.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "UnslothEfficientLoss",
    "unsloth_efficient_ce_loss",
]

import torch
from .utils import *

class UnslothEfficientLoss(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        _input : torch.Tensor,
        weight : torch.Tensor,
        target : torch.Tensor,
        bias  : Optional[torch.Tensor] = None,
        shift : bool = True,
        reduction : str = "mean",
        logit_scale : Optional[float] = None,
        logit_softcapping : Optional[float] = None,
        loss_function : Callable = torch.nn.CrossEntropyLoss,
        ignore_index : int = -100,
        chunk_size : int = 8192,
    ):
        # All Unsloth Studio code licensed under AGPLv3
        device = weight.device
        dtype = _input.dtype
        vocab_size, hd = weight.shape
        _loss_function = loss_function(
            reduction = "sum",
            ignore_index = ignore_index,
        )

        has_grad_weight = weight.requires_grad
        has_grad_input  = _input.requires_grad
        has_grad_bias   = False

        # Mixed precision downcasts from float32 to float16
        weight = weight.to(dtype)
        if bias is not None:
            has_grad_bias = bias.requires_grad
            bias = bias.to(dtype)
        pass

        def process_labels(target):
            if shift:
                shift_target = torch.empty_like(target, device = device, dtype = torch.int64)
                shift_target[..., :-1] = target[..., 1:]
                shift_target[..., -1] = ignore_index
                shift_target = shift_target.view(-1)
            else:
                shift_target = target
            return (
                shift_target,
                (shift_target != ignore_index).sum() if reduction == "mean" else 1.0,
            )
        pass
        if UNSLOTH_COMPILE_ENABLE:
            process_labels = torch.compile(
                process_labels,
                dynamic = None,
                options = torch_compile_options,
            )
            mark_dynamic(target, 1)
        pass
        target, n_labels = process_labels(target)
        if reduction == "sum": n_labels = 1.0

        def compute_loss(input_chunk, weight, bias, target):
            input_chunk = input_chunk.to(weight.device)
            if bias is not None:
                logits = torch.addmm(bias, input_chunk, weight.t())
            else:
                logits = torch.matmul(input_chunk, weight.t())
            
            # Logit Scaling like in Cohere, Granite
            if logit_scale is not None:
                logits = logits * logit_scale

            # Softcapping like in Gemma 2
            if logit_softcapping is not None:
                logits = logits / logit_softcapping
                logits = torch.tanh(logits)
                logits = logits * logit_softcapping
            
            # Upcast and loss_function -> force float32 upcast
            with torch.autocast(device_type = "cuda", enabled = False):
                logits = logits.float()
                loss = _loss_function(logits, target)
            return loss / n_labels
        pass

        grad_weight = torch.zeros_like(weight, dtype = torch.float32, device = device) if has_grad_weight else None
        grad_bias   = torch.zeros_like(bias, dtype = torch.float32, device = device) if has_grad_bias else None
        total_loss  = torch.zeros((), dtype = torch.float32, device = device)
        grad_input  = torch.empty_like(_input, device = device) if has_grad_input else None
        _input      = _input.view(-1, hd)

        # if > 50%, then make a new chunk
        n_chunks = int(round(_input.shape[0] / chunk_size))
        if n_chunks == 0: n_chunks = 1

        def accumulate_chunk(input_chunk, target_chunk, grad_input_chunk):
            chunk_grad_weight = None
            chunk_grad_bias = None
            if has_grad_weight and has_grad_bias and has_grad_input:
                (chunk_grad_input, chunk_grad_weight, chunk_grad_bias,), chunk_loss = torch.func.grad_and_value(
                    compute_loss, argnums = (0, 1, 2,))(
                    input_chunk, weight, bias, target_chunk,
                )
            elif not has_grad_weight and not has_grad_bias and has_grad_input:
                (chunk_grad_input,), chunk_loss = torch.func.grad_and_value(
                    compute_loss, argnums = (0,))(
                    input_chunk, weight, bias, target_chunk,
                )
            elif has_grad_weight and not has_grad_bias and has_grad_input:
                (chunk_grad_input, chunk_grad_weight,), chunk_loss = torch.func.grad_and_value(
                    compute_loss, argnums = (0, 1,))(
                    input_chunk, weight, bias, target_chunk,
                )
            elif not has_grad_weight and has_grad_bias and has_grad_input:
                (chunk_grad_input, chunk_grad_bias,), chunk_loss = torch.func.grad_and_value(
                    compute_loss, argnums = (0, 2,))(
                    input_chunk, weight, bias, target_chunk,
                )
            else:
                raise RuntimeError(
                    f"Unsloth: Loss gradients not supported: {has_grad_weight}, {has_grad_bias}, {has_grad_input}"
                )
                
            if grad_weight      is not None: grad_weight     .add_ (chunk_grad_weight)
            if grad_bias        is not None: grad_bias       .add_ (chunk_grad_bias)
            if grad_input_chunk is not None: grad_input_chunk.copy_(chunk_grad_input)
            total_loss.add_(chunk_loss)
        pass
        if UNSLOTH_COMPILE_ENABLE:
            accumulate_chunk = torch.compile(
                accumulate_chunk,
                dynamic = None,
                options = torch_compile_options,
            )
        pass

        input_chunks  = torch.chunk(_input, n_chunks, dim = 0)
        target_chunks = torch.chunk(target, n_chunks, dim = 0)
        grad_input_chunks = torch.chunk(grad_input.view(-1, hd), n_chunks, dim = 0) \
            if has_grad_input else [None] * n_chunks

        for input_chunk, target_chunk, grad_input_chunk in \
            zip(input_chunks, target_chunks, grad_input_chunks):
            
            if UNSLOTH_COMPILE_ENABLE: 
                mark_dynamic(input_chunk,      0)
                mark_dynamic(target_chunk,     0)
                mark_dynamic(grad_input_chunk, 0)
            accumulate_chunk(input_chunk, target_chunk, grad_input_chunk)
        pass

        ctx.save_for_backward(
            grad_input,
            grad_weight,
            grad_bias,
        )
        return total_loss
    pass

    @staticmethod
    def backward(ctx, grad_output):
        # All Unsloth Studio code licensed under AGPLv3
        (grad_input, grad_weight, grad_bias) = ctx.saved_tensors
        # Multiply by upstream gradients
        if grad_output != 1.0:
            if grad_input  is not None: grad_input .mul_(grad_output)
            if grad_weight is not None: grad_weight.mul_(grad_output)
            if grad_bias   is not None: grad_bias  .mul_(grad_output)
        pass
        return (
            grad_input, grad_weight, None, grad_bias,
            None, None, None, None, None, None, None,
        )
    pass
pass


def unsloth_efficient_ce_loss(
    hidden_states : torch.Tensor,
    lm_head : torch.nn.Linear,
    labels : torch.Tensor,
    shift : bool = True,
    reduction : str = "mean",
    logit_scale : Optional[float] = None,
    logit_softcapping : Optional[float] = None,
    ignore_index : int = -100,
    chunk_size : int = 1024, # Around 512MB per 1024 for 128K vocab
):
    # All Unsloth Studio code licensed under AGPLv3
    assert(type(hidden_states) is torch.Tensor)
    assert(type(lm_head) is torch.nn.Linear)
    assert(type(labels) is torch.Tensor)
    assert(type(shift) is bool)
    assert(reduction == "mean" or reduction == "sum")
    assert(logit_scale is None or type(logit_scale) is float)
    assert(logit_softcapping is None or type(logit_softcapping) is float)
    assert(type(ignore_index) is int)
    assert(type(chunk_size) is int)

    # Dynamic chunk size
    # Smaller ones have less chunks, larger ones more chunks
    vocab_size = lm_head.out_features
    chunk_size = int(chunk_size * ((128 * 1024) / vocab_size))

    return UnslothEfficientLoss.apply(
        hidden_states,
        lm_head.weight,
        labels,
        lm_head.bias,
        shift,
        reduction,
        logit_scale,
        logit_softcapping,
        torch.nn.CrossEntropyLoss,
        ignore_index,
        chunk_size,
    )
pass
